from __future__ import annotations
from ._imports import *

from .button import Button

if t.TYPE_CHECKING:
    from ..projectiles.reaction import Reaction
    from .message import Message


class ButtonReaction(Button, metaclass=abc.ABCMeta):
    """
    An abstract class representing a clickable reaction to a message.
    """

    @ap.async_property
    async def reactions(self) -> t.List["Reaction"]:
        """
        :return: The list of reactions generated by this button. It may vary every time this property is accessed,
                 based on the users who have reacted to the button at the time of access.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def count(self) -> int:
        """
        :return: The count of reactions that this button generated. It may vary every time this property is accessed,
                 based on how many users have reacted to the button at the time of access.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def message(self) -> t.Optional["Message"]:
        """
        :return: The message this button is attached to. Can be :data:`None`, if the button hasn't been attached to a
                 message yet.
        """
        raise exc.NotSupportedError()


__all__ = (
    "ButtonReaction",
)



"""
This module contains the base :class:`.EngineerException` .
"""


class EngineerException(Exception):
    """
    The base class for errors in :mod:`royalnet.engineer`.
    """


__all__ = (
    "EngineerException",
)



import royalnet.royaltyping as t

import sqlalchemy.orm as so
import abc
import async_property as ap

from .. import exc
from ._base import BulletContents


__all__ = (
    "t",
    "so",
    "abc",
    "ap",
    "exc",
    "BulletContents",
)



"""
This module contains :class:`.Conversation`, the base type for all conversations in Royalnet, and
:class:`.DecoratingConversation`, an helper class to instantiate conversations.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import abc
import teleporter as tp

log = logging.getLogger(__name__)


class Conversation:
    """
    :class:`.Conversation`\\ s are objects which can be called to :attr:`.run` a function which awaits
    :class:`~royalnet.engineer.bullet.projectiles._base.Projectile`\\ s incoming from a
    :class:`~royalnet.engineer.sentry.Sentry` .

    Regular coroutine functions can be used instead of this class, at the cost of slightly worse debug information and
    logging.

    .. seealso:: :class:`.DecoratingConversation`
    """

    @abc.abstractmethod
    async def run(self, **kwargs) -> None:
        """
        Run the conversation.

        :param kwargs: The kwargs passed to the :class:`.Conversation` from the
                       :class:`royalnet.engineer.pda.implementations.base.PDAImplementation` .
                       Usually, they include at least ``_sentry``, but more may be available based on
                       the :class:`royalnet.engineer.pda.implementations.base.PDAImplementation`
                       and available :class:`royalnet.engineer.pda.extensions.base.PDAExtension`\\ s.

        :return: :data:`None` to terminate the conversation, or another :class:`.Conversation` to switch to it.
        """
        raise NotImplementedError()

    def __call__(self, **kwargs) -> t.Awaitable[None]:
        log.debug(f"{self}: Called")
        return self.run(**kwargs)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} #{id(self)}>"


class DecoratingConversation(Conversation):
    """
    A decorator-based approach to creating a :class:`.Conversation`.
    """

    def __init__(self, function: t.ConversationProtocol):
        """
        Either pass a :attr:`.function` to this constructor, or use it as a decorator to create a new
        :class:`.Conversation` .

        >>> @DecoratingConversation
        ... async def decoconv(**kwargs):
        ...     ...
            <DecoratedConversation wrapping <function at 0x...>>
        """

        self.function: t.ConversationProtocol = function

    async def run(self, **kwargs) -> None:
        await self.function(**kwargs)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} decorating {self.function}>"


class TeleportingConversation(DecoratingConversation):
    """
    .. todo:: Document this.
    """

    def __init__(self, function: t.ConversationProtocol):
        """
        .. todo:: Document this.
        """

        super().__init__(tp.Teleporter(function, validate_output=False))
        self.bare_function = function

    def __repr__(self):
        return f"<{self.__class__.__qualname__} teleporting {self.bare_function}>"


__all__ = (
    "Conversation",
    "DecoratingConversation",
)



"""
This module contains the base :class:`.Projectile` class.
"""

from __future__ import annotations

import abc

from .. import casing


class Projectile(casing.Casing, metaclass=abc.ABCMeta):
    """
    Abstract base class for external events which can be inserted in a dispenser.
    """


__all__ = (
    "Projectile",
)



from __future__ import annotations
from ._imports import *


class Button(BulletContents, metaclass=abc.ABCMeta):
    """
    An abstract class representing a clickable button.
    """

    @ap.async_property
    async def text(self) -> t.Optional[str]:
        """
        :return: The text displayed on the button.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Button",
)



from __future__ import annotations
from ._imports import *

if t.TYPE_CHECKING:
    from .channel import Channel


class User(BulletContents, metaclass=abc.ABCMeta):
    """
    An abstract class representing a user who can read or send messages in the chat.
    """

    @ap.async_property
    async def name(self) -> t.Optional[str]:
        """
        :return: The user's name.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def database(self, session: so.Session) -> t.Any:
        """
        :param session: A :class:`sqlalchemy.orm.Session` instance to use to fetch the database entry.
        :return: The database entry for this user.
        """
        raise exc.NotSupportedError()

    async def slide(self) -> "Channel":
        """
        Slide into the DMs of the user and get the private channel they share with with the bot.

        :return: The private channel where you can talk to the user.
        """
        raise exc.NotSupportedError()


__all__ = (
    "User",
)



"""
This module contains the base :class:`~royalnet.engineer.pda.extensions.base.PDAExtension`\\ .
"""

import royalnet.royaltyping as t
import abc
import contextlib


class PDAExtension(metaclass=abc.ABCMeta):
    """
    A :class:`.PDAExtension` is an object which extends a
    :class:`~royalnet.engineer.pda.implementations.base.PDAImplementation` by providing additional kwargs to
    :class:`~royalnet.engineer.conversation.Conversation`\\ s.
    """

    @abc.abstractmethod
    @contextlib.asynccontextmanager
    async def kwargs(self, kwargs: t.Kwargs) -> t.Kwargs:
        """
        An :func:`~contextlib.asynccontextmanager` which takes the kwargs that would be passed to the
        :class:`~royalnet.engineer.conversation.Conversation`\\ , modifies them (for example by adding new items) and
        yields them, then performs cleanup operations.
        """

        yield NotImplemented


__all__ = (
    "PDAExtension",
)



"""
This module contains the base :class:`.PDA` class.
"""

import asyncio
import royalnet.royaltyping as t
import logging

if t.TYPE_CHECKING:
    from royalnet.engineer.pda.implementations.base import PDAImplementation
    DispenserKey = t.TypeVar("DispenserKey")

log = logging.getLogger(__name__)


class PDA:
    """
    .. todo:: Document this.
    """

    def __init__(self, implementations: list["PDAImplementation"]):
        self.implementations: dict[str, "PDAImplementation"] = {}
        for implementation in implementations:
            implementation.bind(pda=self)
            self.implementations[implementation.name] = implementation

    def __repr__(self):
        return f"<{self.__class__.__qualname__} implementing {', '.join(self.implementations.keys())}>"

    def __len__(self):
        return len(self.implementations)

    async def _run(self):
        log.info("Running all implementations...")
        await asyncio.gather(*[implementation.run() for implementation in self.implementations.values()])
        log.fatal("All implementations have finished running?!")

    def run(self):
        log.debug("Running blockingly all implementations...")
        asyncio.run(self._run())
        log.fatal("Blocking call has finished?!")


__all__ = (
    "PDA",
)



"""
Base models are :class:`pydantic.BaseModel` that have specific configuration tailored for certain uses.
"""

import pydantic


class RoyalnetModel(pydantic.BaseModel):
    """
    A model for generic data.
    """

    class Config(pydantic.BaseModel.Config):
        """
        The default :mod:`pydantic` configuration.

        .. seealso:: `Pydantic Configuration <https://pydantic-docs.helpmanual.io/usage/model_config/>`_,
                     :class:`pydantic.BaseModel.Config`
        """
        pass


class OrmModel(RoyalnetModel):
    """
    A model for :mod:`sqlalchemy` table data.
    """

    class Config(RoyalnetModel.Config):
        """
        A configuration which allows for the loading of data from ``__getattr__`` instead of ``__getitem__``.

        .. seealso:: `Pydantic ORM Mode <https://pydantic-docs.helpmanual.io/usage/models/#orm-mode-aka-arbitrary
                     -class-instances>`_
        """
        orm_mode = True


__all__ = (
    "RoyalnetModel",
    "OrmModel",
)



"""
This module contains the :class:`.Teleporter` class and its exceptions.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import pydantic
import inspect

from . import exc

Value = t.TypeVar("Value")
log = logging.getLogger(__name__)


class TeleporterError(exc.EngineerException, pydantic.ValidationError):
    """
    The base class for errors in :mod:`royalnet.engineer.teleporter`.
    """


class InTeleporterError(TeleporterError):
    """
    The input parameters validation failed.
    """


class OutTeleporterError(TeleporterError):
    """
    The return value validation failed.
    """


class Teleporter:
    """
    A :class:`.Teleporter` is a function wrapper which uses :mod:`pydantic` to perform type checking
    """

    def __init__(self,
                 f: t.Callable[..., t.Any],
                 validate_input: bool = True,
                 validate_output: bool = True):
        self.f: t.Callable = f
        """
        The function which is having its parameters and return value validated.
        """

        self.InputModel: t.Type[pydantic.BaseModel] = self._create_input_model() if validate_input else None
        """
        The :mod:`pydantic` model used to validate input parameters.
        """

        self.OutputModel: t.Type[pydantic.BaseModel] = self._create_output_model() if validate_output else None
        """
        The :mod:`pydantic` model used to validate the return value.
        """

    def __repr__(self):
        if self.InputModel and self.OutputModel:
            validation = "validating input and output"
        elif self.InputModel:
            validation = "validating only input"
        elif self.OutputModel:
            validation = "validating only output"
        else:
            validation = "not validating anything"
        return f"<{self.__class__.__qualname__} {validation}>"

    @staticmethod
    def _parameter_to_field(param: inspect.Parameter, **kwargs) -> t.Tuple[type, pydantic.fields.FieldInfo]:
        """
        Convert a :class:`inspect.Parameter` to a type-field :class:`tuple`, which can be easily passed to
        :func:`pydantic.create_model`.

        If the parameter is already a :class:`pydantic.FieldInfo` (created by :func:`pydantic.Field`), it will be
        returned as the value, without creating a new model.

        :param param: The :class:`inspect.Parameter` to convert.
        :param kwargs: Additional kwargs to pass to the field.
        :return: A :class:`tuple`, where the first element is a :class:`type` and the second is a
                 :class:`pydantic.Field`.
        """
        if isinstance(param.default, pydantic.fields.FieldInfo):
            log.debug(f"Parameter {param} is a pydantic.Field, leaving it untouched...")
            return (
                param.annotation,
                param.default
            )
        else:
            log.debug(f"Parameter {param} is not a pydantic.Field, converting it to one...")
            return (
                param.annotation,
                pydantic.Field(
                    default=param.default if param.default is not inspect.Parameter.empty else ...,
                    title=param.name,
                    **kwargs,
                ),
            )

    class TeleporterDefaultConfig(pydantic.BaseConfig):
        """
        A :mod:`pydantic` model config which allows for arbitrary types.
        """
        arbitrary_types_allowed = True

    def get_model_config(self):
        """
        Get the :mod:`pydantic` config to use in both input and output, if :meth:`.get_input_model_config` and
        :meth:`.get_output_model_config` are not overridden.

        :return: A :mod:`pydantic` config.
        """
        log.debug("Getting default model config...")
        return self.TeleporterDefaultConfig

    def get_input_model_config(self):
        """
        Get the :mod:`pydantic` config to use while creating input models.

        :return: A :mod:`pydantic` config.
        """
        log.debug("Getting common model config...")
        return self.get_model_config()

    def get_output_model_config(self):
        """
        Get the :mod:`pydantic` config to use while creating output models.

        :return: A :mod:`pydantic` config.
        """
        log.debug("Getting common model config...")
        return self.get_model_config()

    def _create_input_model(self,
                            **extra_fields) -> t.Type[pydantic.BaseModel]:
        """
        Create a pydantic model based on the arguments of the :attr:`f` function.

        Arguments starting with ``_`` are ignored.

        The model is created using the config obtained through :meth:`.get_input_model_config` .

        :param extra_fields: Extra fields to be added to the model.
        :return: The created pydantic model.
        """
        log.debug(f"Getting function signature of: {self.f!r}")
        signature: inspect.Signature = inspect.signature(self.f)

        log.debug(f"Converting parameter annotations of {self.f!r} to fields...")
        fields = {
            key: self._parameter_to_field(value)
            for key, value in signature.parameters.items()
            if not key.startswith("_")
        }

        log.debug(f"Creating input model with parsed fields {fields!r} and extra fields {extra_fields!r}...")
        return pydantic.create_model(
            f"{self.__class__.__name__}InputModel",
            __config__=self.get_input_model_config(),
            **fields,
            **extra_fields
        )

    def _create_output_model(self) -> t.Type[pydantic.BaseModel]:
        """
        Create a pydantic model based on the return value of the :attr:`f` function.

        The model is created using the config obtained through :meth:`.get_output_model_config` .

        :return: The created pydantic model.
        """
        log.debug(f"Getting function signature of: {self.f!r}")
        signature: inspect.Signature = inspect.signature(self.f)

        log.debug(f"Creating output model...")
        return pydantic.create_model(
            f"{self.__class__.__name__}OutputModel",
            __config__=self.get_output_model_config(),
            __root__=(signature.return_annotation, pydantic.Field(..., title="Returns"))
        )

    def teleport_in(self, **kwargs) -> pydantic.BaseModel:
        """
        Instantiate the :attr:`.InputModel` with the passed kwargs.

        :param kwargs: The keyword arguments that should be passed to the model.
        :return: The created model.
        :raises .InTeleporterError: If the kwargs fail the validation.
        """
        log.debug(f"Teleporting in: {kwargs!r}")
        try:
            return self.InputModel(**kwargs)
        except pydantic.ValidationError as e:
            log.error(f"Teleport in failed: {e!r}")
            raise InTeleporterError(errors=e.raw_errors, model=e.model)

    def teleport_out(self, value: Value) -> pydantic.BaseModel:
        """
        Instantiate the :attr:`.OutputModel` with the passed value.

        :param value: The value that should be validated.
        :return: The created model.
        :raises .OutTeleporterError: If the value fails the validation.
        """
        log.debug(f"Teleporting out: {value!r}")
        try:
            return self.OutputModel(__root__=value)
        except pydantic.ValidationError as e:
            log.error(f"Teleport out failed: {e!r}")
            raise OutTeleporterError(errors=e.raw_errors, model=e.model)

    @staticmethod
    def _split_kwargs(**kwargs) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        """
        Split the passed kwargs in two different :class:`dict`:
        - One containing the arguments that **do not start with ``_``**;
        - Another containing the ones which do.

        :return: A tuple of :class:`dict`, where the second contains the ones starting with ``_``, and the first
                 contains the rest.
        """
        model_params = {}
        extra_params = {}
        for key, value in kwargs.items():
            if key.startswith("_"):
                log.debug(f"Found extra keyword argument: {key}")
                extra_params[key] = value
            else:
                log.debug(f"Found model keyword argument: {key}")
                model_params[key] = value
        return model_params, extra_params

    def _run(self, **kwargs) -> t.Any:
        """
        Run the :class:`.Teleporter` synchronously.
        """
        if self.InputModel:
            log.debug("Validating input...")
            model_kwargs, extra_kwargs = self._split_kwargs(**kwargs)
            model_kwargs = self.teleport_in(**kwargs).dict()
            kwargs = {**model_kwargs, **extra_kwargs}
        result = self.f(**kwargs)
        if self.OutputModel:
            result = self.teleport_out(result).__root__
        return result

    async def _run_async(self, **kwargs) -> t.Awaitable[t.Any]:
        """
        Run the :class:`.Teleporter` asynchronously.
        """
        if self.InputModel:
            log.debug("Validating input...")
            model_kwargs, extra_kwargs = self._split_kwargs(**kwargs)
            model_kwargs = self.teleport_in(**kwargs).dict()
            kwargs = {**model_kwargs, **extra_kwargs}
        result = await self.f(**kwargs)
        if self.OutputModel:
            result = self.teleport_out(result).__root__
        return result

    def __call__(self, **kwargs) -> t.Any:
        if inspect.iscoroutinefunction(self.f):
            return self._run_async(**kwargs)
        else:
            return self._run(**kwargs)


__all__ = (
    "InTeleporterError",
    "OutTeleporterError",
    "Teleporter",
    "TeleporterError",
)



"""
This module contains the projectiles related to messages: :class:`.MessageReceived`, :class:`MessageEdited` and
:class:`MessageDeleted`.
"""

from __future__ import annotations
from ._imports import *

if t.TYPE_CHECKING:
    from ..contents.message import Message


class MessageReceived(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing the reception of a single message.
    """

    @ap.async_property
    async def message(self) -> "Message":
        """
        :return: The received Message.
        """
        raise exc.NotSupportedError()


class MessageEdited(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing the editing of a single message.
    """

    @ap.async_property
    async def message(self) -> "Message":
        """
        :return: The edited Message.
        """
        raise exc.NotSupportedError()


class MessageDeleted(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing the deletion of a single message.
    """

    @ap.async_property
    async def message(self) -> "Message":
        """
        :return: The edited Message.
        """
        raise exc.NotSupportedError()


__all__ = (
    "MessageReceived",
    "MessageEdited",
    "MessageDeleted",
)



"""
This module contains the :class:`.Discard` special exception.
"""


class Discard(BaseException):
    """
    A special exception which should be raised by :class:`~royalnet.engineer.wrench.Wrench`\\ es if a certain object
    should be discarded from the queue.
    """

    def __init__(self, obj, message):
        self.obj = obj
        self.message = message

    def __repr__(self):
        return f"<{self.__class__.__qualname__}>"

    def __str__(self):
        return f"Discarded {self.obj}: {self.message}"


__all__ = (
    "Discard",
)



"""
This module contains the base :class:`.PDAImplementation` and its basic implementation
:class:`.ConversationListImplementation` .
"""

import royalnet.royaltyping as t
import abc
import contextlib
import asyncio
import logging
from royalnet.engineer.dispenser import Dispenser

if t.TYPE_CHECKING:
    from royalnet.engineer.pda.extensions.base import PDAExtension
    from royalnet.engineer.pda.base import PDA
    from royalnet.engineer.bullet.projectiles import Projectile

DispenserKey = t.Hashable


class PDAImplementation(metaclass=abc.ABCMeta):
    """
    .. todo:: Document this.
    """

    def __init__(self, name: str, extensions: list["PDAExtension"] = None):
        self.name: str = f"{self.namespace}.{name}"
        """
        .. todo:: Document this.
        """

        self.extensions: list["PDAExtension"] = extensions or []
        """
        .. todo:: Document this.
        """

        self.bound_to: t.Optional["PDA"] = None
        """
        .. todo:: Document this.
        """

        self.logger_name: str = f"{__name__}.PDAImplementation.{self.namespace}.{name}"
        """
        .. todo:: Document this.
        """

        self.log: logging.Logger = logging.getLogger(self.logger_name)
        """
        .. todo:: Document this.
        """

    def __repr__(self):
        return f"<PDAImplementation {self.name}>"

    def __str__(self):
        return self.name

    def bind(self, pda: "PDA") -> None:
        """
        .. todo:: Document this.
        """

        self.log.debug(f"Trying to bind to {pda!r}...")
        if self.bound_to is not None:
            self.log.error(f"Already bound to {pda!r}!")
            raise ImplementationAlreadyBoundError()

        self.bound_to = pda
        self.log.info(f"Bound to {pda!r}!")

    @property
    @abc.abstractmethod
    def namespace(self):
        """
        .. todo:: Document this.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    async def run(self):
        """
        .. todo:: Document this.
        """

        raise NotImplementedError()


class ImplementationException(Exception):
    """
    .. todo:: Document this.
    """


class ImplementationAlreadyBoundError(ImplementationException):
    """
    .. todo:: Document this.
    """


class ConversationListImplementation(PDAImplementation, metaclass=abc.ABCMeta):
    """
    A :class:`.PDAImplementation` which instantiates runs all the 
    :class:`~royalnet.engineer.conversation.Conversation`\\ s in the :attr:`.conversations` :class:`list` 
    before :meth:`put`\\ ting a :class:`~royalnet.engineer.bullet.projectile.Projectile` in a
    :class:`~royalnet.engineer.dispenser.Dispenser` .
    """

    def __init__(self, name: str, extensions: list["PDAExtension"] = None):
        super().__init__(name=name, extensions=extensions)

        self.conversations: list[t.ConversationProtocol] = self._create_conversations()
        """
        A :class:`list` of :class:`~royalnet.engi.conversation.Conversation`\\ s that should be run before 
        :meth:`put`\\ ting a :class:`~royalnet.engineer.bullet.projectile.Projectile` in a
        :class:`~royalnet.engineer.dispenser.Dispenser` .
        """

        self.dispensers: dict[DispenserKey, "Dispenser"] = self._create_dispensers()
        """
        A :class:`dict` which maps :func:`hash`\\ able objects to a :class:`~royalnet.engineer.dispenser.Dispenser` .
        """

    def _create_conversations(self) -> list[t.ConversationProtocol]:
        """
        Create the :attr:`.conversations` :class:`list` of the :class:`.ConversationListPDA`\\ .

        :return: The created :class:`list`, empty by default.
        """

        self.log.debug(f"Creating conversations list...")
        return []

    def _create_dispensers(self) -> dict[t.Any, "Dispenser"]:
        """
        Create the :attr:`.dispensers` dictionary of the PDA.

        :return: The created :class:`dict`, empty by default.
        """

        self.log.debug(f"Creating dispensers list...")
        return {}

    def get(self, key: DispenserKey) -> t.Optional["Dispenser"]:
        """
        Get a :class:`~royalnet.engineer.dispenser.Dispenser` from :attr:`.dispensers` given an :func:`hash`\\ able 
        object.

        :param key: The key to get the dispenser with.
        :return: The retrieved dispenser.

        .. seealso:: :meth:`dict.get`
        """

        self.log.debug(f"Getting dispenser with key: {key!r}")
        return self.dispensers.get(key)

    def _create_dispenser(self) -> "Dispenser":
        """
        Create a new :class:`~royalnet.engineer.dispenser.Dispenser` .

        :return: The created :class:`~royalnet.engineer.dispenser.Dispenser` .
        """

        self.log.debug(f"Creating new dispenser...")
        return Dispenser()

    def get_or_create_dispenser(self, key: DispenserKey) -> "Dispenser":
        """
        Try to :meth:`.get` the :class:`~royalnet.engineer.dispenser.Dispenser` with the specified key, or
        :meth:`._create_dispenser` a new one if it isn't available.

        :param key: The key of the :class:`~royalnet.engineer.dispenser.Dispenser` .
        :return: The retrieved or created :class:`~royalnet.engineer.dispenser.Dispenser` .
        """

        if key not in self.dispensers:
            self.log.debug(f"{self!r}: Dispenser {key!r} does not exist, creating a new one...")
            self.dispensers[key] = self._create_dispenser()
        return self.get(key=key)

    @contextlib.asynccontextmanager
    async def kwargs(self, conv: t.ConversationProtocol) -> t.Kwargs:
        """
        :func:`contextlib.asynccontextmanager` factory which yields the arguments to pass to newly created
        :class:`~royalnet.engineer.conversation.Conversation`\\ s .

        By default, the following arguments are passed:
        - ``_pda``: contains the :class:`.PDA` this implementation is bound to.
        - ``_imp``: contains this :class:`.PDAImplementation` .
        - ``_conv``: contains the :class:`~royalnet.engineer.conversation.Conversation` which was just created.

        :param conv: The :class:`~royalnet.engineer.conversation.Conversation` to create the args for.
        :return: The corresponding :func:`contextlib.asynccontextmanager`\\ .
        """
    
        self.log.debug(f"Creating kwargs for: {conv!r}")

        default_kwargs = {
            "_pda": self.bound_to,
            "_imp": self,
            "_conv": conv,
        }

        async with self._kwargs(default_kwargs, self.extensions) as kwargs:
            self.log.info(f"Yielding kwargs for {conv!r}: {kwargs!r}")
            yield kwargs

    @contextlib.asynccontextmanager
    async def _kwargs(self, kwargs: t.Kwargs, remaining: list["PDAExtension"]) -> t.Kwargs:
        """
        :func:`contextlib.asynccontextmanager` factory used internally to recurse the generation and cleanup of
        :meth:`kwargs` .

        :param kwargs: The current ``kwargs`` .
        :param remaining: The extensions that haven't been processed yet.
        :return: The corresponding :func:`contextlib.asynccontextmanager`\\ .
        """

        if len(remaining) == 0:
            self.log.debug(f"Kwargs recursion ended!")
            yield kwargs
        else:
            extension = remaining[0]
            self.log.debug(f"Getting kwargs from {extension}, {len(remaining)} left...")
            async with extension.kwargs(kwargs) as kwargs:
                self.log.debug(f"Recursing...")
                async with self._kwargs(kwargs=kwargs, remaining=remaining[1:]) as kwargs:
                    self.log.debug(f"Bubbling up yields...")
                    yield kwargs

    def register_conversation(self, conversation: t.ConversationProtocol) -> None:
        """
        Register a new :class:`~royalnet.engineer.conversation.Conversation` to be run when a new
        :class:`~royalnet.engineer.bullet.projectile.Projectile` is :meth:`.put`\\ .

        :param conversation: The :class:`~royalnet.engineer.conversation.Conversation` to register.
        """

        self.log.debug(f"Registering: {conversation!r}")
        self.conversations.append(conversation)

    def unregister_conversation(self, conversation: t.ConversationProtocol) -> None:
        """
        Unregister a :class:`~royalnet.engineer.conversation.Conversation`, stopping it from being run when a new
        :class:`~royalnet.engineer.bullet.projectile.Projectile` is :meth:`.put`\\ .

        :param conversation: The :class:`~royalnet.engineer.conversation.Conversation` to unregister.
        """

        self.log.debug(f"Unregistering: {conversation!r}")
        self.conversations.remove(conversation)

    async def _run_conversation(self, dispenser: "Dispenser", conv: t.ConversationProtocol) -> None:
        """
        Run the passed :class:`~royalnet.engineer.conversation.Conversation` in the passed
        :class:`~royalnet.engineer.dispenser.Dispenser`\\ , while passing the :meth:`.kwargs` provided by the
        :class:`.PDA` .

        :param dispenser: The :class:`~royalnet.engineer.dispenser.Dispenser` to run the
                          :class:`~royalnet.engineer.conversation.Conversation` in.
        :param conv: The :class:`~royalnet.engineer.conversation.Conversation` to run.
        """

        try:
            async with self.kwargs(conv=conv) as kwargs:
                self.log.debug(f"Running {conv!r} in {dispenser!r}...")
                await dispenser.run(conv=conv, **kwargs)
        except Exception as exception:
            try:
                await self._handle_conversation_exc(dispenser=dispenser, conv=conv, exception=exception)
            except Exception as exception:
                self.log.error(f"Failed to handle conversation exception: {exception!r}")

    async def _handle_conversation_exc(
            self,
            dispenser: "Dispenser",
            conv: t.ConversationProtocol,
            exception: Exception
    ) -> None:
        """
        Handle :exc:`Exception`\\ s that were not caught by :class:`~royalnet.engineer.conversation.Conversation`\\ s.

        :param dispenser: The dispenser which hosted the :class:`~royalnet.engineer.conversation.Conversation`\\ .
        :param conv: The :class:`~royalnet.engineer.conversation.Conversation` which didn't catch the error.
        :param exception: The :class:`Exception` that was raised.
        """

        self.log.error(f"Unhandled {exception} in {conv} run in {dispenser}!")

    async def _schedule_conversations(self, dispenser: "Dispenser") -> list[asyncio.Task]:
        """
        Schedule the execution of instance of all the :class:`~royalnet.engineer.conversation.Conversation`\\ s listed
        in :attr:`.conversations` in the specified :class:`~royalnet.engineer.dispenser.Dispenser`\\ .

        :param dispenser: The :class:`~royalnet.engineer.dispenser.Dispenser` to run the
                          :class:`~royalnet.engineer.conversation.Conversation`\\ s in.
        :return: The :class:`list` of :class:`asyncio.Task`\\ s that were created.

        .. seealso:: :meth:`._run_conversation`
        """

        if dispenser.locked_by:
            self.log.debug("Refusing to run new Conversations in a locked Dispenser")
            return []

        self.log.info(f"Running in {dispenser!r} all conversations...")

        tasks: list[asyncio.Task] = []
        for conv in self.conversations:

            self.log.debug(f"Creating task for: {conv!r}")
            task = asyncio.create_task(self._run_conversation(dispenser=dispenser, conv=conv))

            tasks.append(task)

        self.log.debug(f"Running a event loop cycle...")
        await asyncio.sleep(0)

        self.log.info(f"Tasks created: {tasks!r}")
        return tasks

    async def put(self, key: DispenserKey, projectile: "Projectile") -> None:
        """
        Put a :class:`~royalnet.engineer.bullet.projectile.Projectile` in the
        :class:`~royalnet.engineer.dispenser.Dispenser` with the specified key.

        :param key: The key identifying the :class:`~royalnet.engineer.dispenser.Dispenser` among the other
                    :attr:`.dispensers`.
        :param projectile: The :class:`~royalnet.engineer.bullet.projectile.Projectile` to insert.
        """

        self.log.debug(f"Finding dispenser {key!r} to put {projectile!r} in...")
        dispenser = self.get_or_create_dispenser(key=key)

        self.log.debug(f"Running all conversations...")
        await self._schedule_conversations(dispenser=dispenser)

        self.log.debug(f"Putting {projectile!r} in {dispenser!r}...")
        await dispenser.put(projectile)

        self.log.debug(f"Running a event loop cycle...")
        await asyncio.sleep(0)


__all__ = (
    "PDAImplementation",
    "ImplementationException",
    "ImplementationAlreadyBoundError",
    "ConversationListImplementation",
)



"""
This modules contains the exceptions which can happen in bullets.
"""

from .. import exc


class BulletException(exc.EngineerException):
    """
    The base class for errors in :mod:`royalnet.engineer.bullet`.
    """


class FrontendError(BulletException):
    """
    An error occoured while performing a frontend operation, such as sending a message.
    """


class NotSupportedError(FrontendError, NotImplementedError):
    """
    The requested property isn't available on the current implementation.
    """


class ForbiddenError(FrontendError):
    """
    The bot does not have sufficient permissions to perform an operation.
    """



"""
Chat bot utilities.

All names are inspired by the `Engineer Class of Team Fortress 2 <https://wiki.teamfortress.com/wiki/Engineer>`_.

.. image: /_static/engineer_diagram.png
"""

from .command import *
from .conversation import *
from .discard import *
from .dispenser import *
from .exc import *
from .sentry import *
from .teleporter import *
from .wrench import *
from .bullet import *
from .pda import *



"""
This module contains the :class:`.Dispenser` class.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import contextlib

from .sentry import SentrySource
from .exc import EngineerException
from .bullet.projectiles import Projectile

log = logging.getLogger(__name__)


class DispenserException(EngineerException):
    """
    The base class for errors in :mod:`royalnet.engineer.dispenser`\\ .
    """


class LockedDispenserError(DispenserException):
    """
    The :class:`.Dispenser` couldn't start a new :class:`~royalnet.engineer.conversation.Conversation` as it is
    currently :attr:`.Dispenser.lock`\\ ed.
    """
    def __init__(self, locked_by, *args):
        super().__init__(*args)
        self.locked_by = locked_by


class Dispenser:
    """
    A :class:`.Dispenser` is an object which instantiates multiple :class:`~royalnet.engineer.sentry.Sentry` and
    multiplexes and distributes incoming :class:`~royalnet.engineer.bullet.projectiles._base.Projectile`\\ s to them.

    They usually represent a single "conversation channel" with the bot: either a chat channel, or an user.
    """

    def __init__(self):
        self.sentries: t.List[SentrySource] = []
        """
        A :class:`list` of all the running sentries of this dispenser.
        """

        self.locked_by: t.List[t.ConversationProtocol] = []
        """
        The conversation that is currently locking this dispenser.
        
        .. seealso:: :meth:`.lock`
        """

    async def put(self, item: Projectile) -> None:
        """
        Insert a new :class:`~royalnet.engineer.bullet.projectiles._base.Projectile` in the queues of all the
        running :attr:`.sentries`.

        :param item: The :class:`~royalnet.engineer.bullet.projectiles._base.Projectile` to insert.
        """
        log.debug(f"Putting {item!r}...")
        for sentry in self.sentries:
            await sentry.put(item)

    @contextlib.contextmanager
    def sentry(self, *args, **kwargs):
        """
        A :func:`~contextlib.contextmanager` which creates a :class:`.SentrySource` and keeps it in :attr:`.sentries`
        while it is being used.
        """
        log.debug("Creating a new SentrySource...")
        sentry = SentrySource(dispenser=self, *args, **kwargs)

        log.debug(f"Adding: {sentry!r}")
        self.sentries.append(sentry)

        log.debug(f"Yielding: {sentry!r}")
        yield sentry

        log.debug(f"Removing from the sentries list: {sentry!r}")
        self.sentries.remove(sentry)

    async def run(self, conv: t.ConversationProtocol, **kwargs) -> None:
        """
        Run a :class:`~royalnet.engineer.conversation.Conversation`\\ .

        :param conv: The :class:`~royalnet.engineer.conversation.Conversation` to run.
        :raises .LockedDispenserError: If the dispenser is currently :attr:`.locked_by` a :class:`.Conversation`.
        """
        log.debug(f"Trying to run: {conv!r}")

        if self.locked_by:
            log.debug(f"Dispenser is locked by {self.locked_by!r}, refusing to run {conv!r}")
            raise LockedDispenserError(
                f"The Dispenser is currently locked and cannot start any new Conversation.", self.locked_by)

        log.debug(f"Running: {conv!r}")
        with self.sentry() as sentry:
            await conv(_sentry=sentry, **kwargs)

    @contextlib.contextmanager
    def lock(self, conv: t.ConversationProtocol):
        """
        Lock the :class:`.Dispenser` while this :func:`~contextlib.contextmanager` is in scope.

        A locked :class:`.Dispenser` will refuse to :meth:`.run` any new conversations,
        raising :exc:`.LockedDispenserError` instead.

        :param conv: The conversation that requested the lock.

        .. seealso:: :attr:`.locked_by`
        """
        log.debug(f"Adding lock: {conv!r}")
        self.locked_by.append(conv)

        try:
            yield
        finally:
            log.debug(f"Clearing lock: {conv!r}")
            self.locked_by.remove(conv)


__all__ = (
    "Dispenser",
    "DispenserException",
    "LockedDispenserError",
)



import pytest
from royalnet.lazy import Lazy


def test_single_eval_no_args():
    lazy = Lazy(lambda: 1)
    assert not lazy.evaluated
    assert lazy._result is None
    result = lazy.evaluate()
    assert result == 1
    assert lazy.evaluated
    assert lazy._result == 1


def test_single_eval_with_args():
    lazy = Lazy(lambda a, b: a + b, 10, 10)
    result = lazy.evaluate()
    assert result == 20
    assert lazy.evaluated
    assert lazy._result == 20


def test_chained_eval():
    twenty = Lazy(lambda a, b: a + b, 10, 10)
    assert not twenty.evaluated
    assert twenty._result is None
    twentyfive = Lazy(lambda a, b: a + b, twenty, 5)
    assert not twentyfive.evaluated
    assert twentyfive._result is None
    result = twentyfive.evaluate()
    assert result == 25
    assert twenty.evaluated
    assert twenty._result == 20
    assert twentyfive.evaluated
    assert twentyfive._result == 25



"""
This module contains the :class:`.Reaction` projectile.
"""

from __future__ import annotations
from ._imports import *

if t.TYPE_CHECKING:
    from ..contents.user import User
    from ..contents.button_reaction import ButtonReaction


class Reaction(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing a reaction of a single user to a message, generated by clicking on a ButtonReaction.
    """

    @ap.async_property
    async def user(self) -> "User":
        """
        :return: The user who reacted to the message.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def button(self) -> "ButtonReaction":
        """
        :return: The ButtonReaction that the user pressed to generate this reaction.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Reaction",
)



"""
.. todo:: Document this.
"""

import royalnet.royaltyping as t
import royalnet.engineer.conversation as c
import royalnet.engineer.sentry as s
import royalnet.engineer.bullet as b
import abc
import logging

log = logging.getLogger(__name__)


class Router(c.Conversation, metaclass=abc.ABCMeta):
    """
    .. todo:: Document this.
    """

    def __init__(self):
        """
        .. todo:: Document this.
        """

        self.by_pattern: dict[t.Pattern, t.ConversationProtocol] = {}
        self.by_name: dict[str, t.ConversationProtocol] = {}
        self.by_command: dict[t.ConversationProtocol, t.List[str]] = {}

    def register_conversation(self, conv: t.ConversationProtocol, names: t.List[str], patterns: t.List[t.Pattern]):
        """
        .. todo:: Document this.
        """

        log.debug(f"Registering {conv!r}...")

        log.debug("Names:")
        for name in names:
            log.debug(f"{name!r} → {conv!r}")
            self.by_name[name] = conv
        log.debug(f"{conv!r} → {names!r}")

        log.debug("Patterns:")
        for pattern in patterns:
            log.debug(f"{pattern.pattern!r} → {conv!r}")
            self.by_pattern[pattern] = conv

    async def run(self, _sentry: s.Sentry, **kwargs) -> None:
        """
        .. todo:: Document this.
        """

        dispenser = _sentry.dispenser()

        log.debug(f"Locking {dispenser!r}...")
        with _sentry.dispenser().lock(self):

            log.debug(f"Awaiting a projectile...")
            projectile: b.Projectile = await _sentry

            log.debug(f"Received: {projectile!r}")

            log.debug(f"Ensuring the projectile is a MessageReceived: {projectile!r}")
            if not isinstance(projectile, b.MessageReceived):
                log.debug(f"Returning: {projectile!r} is not a message")
                return

            log.debug(f"Getting message of: {projectile!r}")
            if not (msg := await projectile.message):
                log.warning(f"Returning: {projectile!r} has no message")
                return

            log.debug(f"Getting message text of: {msg!r}")
            if not (text := await msg.text):
                log.debug(f"Returning: {msg!r} has no text")
                return

            log.debug(f"Message text is: {text!r}")

            for pattern in self.by_pattern.keys():
                log.debug(f"Trying to search for {pattern!r} in the text")
                match = pattern.search(text)
                if match:
                    conversation = self.by_pattern[pattern]
                    log.debug(f"Matched with {pattern!r}, running conversation {conversation}")
                    await conversation(_sentry=_sentry, **kwargs, _conv=conversation, _router=self)
                    return
            else:
                log.debug("No matches found")



"""
This module contains the :class:`~royalnet.engineer.pda.extensions.base.PDAExtension`\\ s that allow
:class:`~royalnet.engineer.conversation.Conversation`\\ s to access a database.
"""

import royalnet.royaltyping as t
import sqlalchemy
import sqlalchemy.orm
import contextlib
import logging

from . import base

log = logging.getLogger(__name__)


class SQLAlchemyExtension(base.PDAExtension):
    """
    Extends a :class:`~royalnet.engineer.pda.implementations.base.PDAImplementation` by adding a :mod:`sqlalchemy`
    session to conversations through the ``_session`` kwarg.
    """

    def __init__(self, engine: sqlalchemy.engine.Engine, session_kwargs: t.Kwargs = None, kwarg_name: str = "_session"):
        super().__init__()
        self.engine: sqlalchemy.engine.Engine = engine
        """
        The :class:`sqlalchemy.engine.Engine` to use.
        """

        self.Session: sqlalchemy.orm.sessionmaker = sqlalchemy.orm.sessionmaker(bind=self.engine)
        """
        The :class:`sqlalchemy.orm.sessionmaker` to use when creating new sessions.
        """

        self.session_kwargs: t.Kwargs = {"future": True, **(session_kwargs or {})}
        """
        Additional kwargs to be passed to the :class:`sqlalchemy.orm.sessionmaker` when instantiating a new Session.
        
        Defaults to ``{"future": True}`` .
        """

        self.kwarg_name: str = kwarg_name
        """
        The name of the kwarg to add.
        
        Defaults to ``"_session"``.
        """

    def __repr__(self):
        return f"<{self.__class__.__qualname__} with engine {self.engine}>"

    @contextlib.asynccontextmanager
    async def kwargs(self, kwargs: t.Kwargs) -> t.Kwargs:
        log.debug(f"{self!r}: Creating session...")
        with self.Session(**self.session_kwargs) as session:
            log.debug(f"{self!r}: Yielding kwargs...")
            yield {
                **kwargs,
                self.kwarg_name: session,
            }
            log.debug(f"{self!r}: Closing session...")
        log.debug(f"{self!r}: Session closed!")


__all__ = (
    "SQLAlchemyExtension",
)



from __future__ import annotations
from royalnet.royaltyping import *


Result = TypeVar("Result")


class Lazy:
    def __init__(self, obj: Callable[[Any], Result], *args, **kwargs):
        self._func: Callable[[Any], Result] = obj
        self._args = args
        self._kwargs = kwargs
        self.evaluated = False
        self._result = None

    def _evaluate_args(self) -> List[Any]:
        evaluated_args = []
        for arg in self._args:
            if isinstance(arg, Lazy):
                arg = arg.evaluate()
            evaluated_args.append(arg)
        return evaluated_args

    def _evaluate_kwargs(self) -> Dict[str, Any]:
        evaluated_kwargs = {}
        for key, value in self._kwargs.items():
            if isinstance(value, Lazy):
                value = value.evaluate()
            evaluated_kwargs[key] = value
        return evaluated_kwargs

    def evaluate(self, *args, **kwargs) -> Result:
        if not self.evaluated:
            self._result = self._func(*self._evaluate_args(), *args, **self._evaluate_kwargs(), **kwargs)
            self.evaluated = True
        return self._result

    @property
    def e(self) -> Result:
        return self.evaluate()

    @classmethod
    def decorator(cls, *args, **kwargs):
        def deco(func):
            return cls(func, *args, **kwargs)
        return deco


__all__ = (
    "Lazy",
)



"""
This module contains the projectiles related to user actions, such as :class:`.UserJoined`, :class:`.UserLeft` and
:class:`.UserUpdate`.
"""

from __future__ import annotations
from ._imports import *

if t.TYPE_CHECKING:
    from ..contents.user import User


class UserJoined(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing an user who just joined the chat channel.
    """

    @ap.async_property
    async def user(self) -> "User":
        """
        :return: The user who joined.
        """
        raise exc.NotSupportedError()


class UserLeft(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing an user who just left the chat channel.
    """

    @ap.async_property
    async def user(self) -> "User":
        """
        :return: The user who left.
        """
        raise exc.NotSupportedError()


class UserUpdate(Projectile, metaclass=abc.ABCMeta):
    """
    An abstract class representing a change in status of an user in the chat channel.
    """

    @ap.async_property
    async def user(self) -> "User":
        """
        :return: The user who joined.
        """
        raise exc.NotSupportedError()


__all__ = (
    "UserJoined",
    "UserLeft",
    "UserUpdate",
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['royalnet',
 'royalnet.alchemist',
 'royalnet.engineer',
 'royalnet.engineer.bullet',
 'royalnet.engineer.bullet.contents',
 'royalnet.engineer.bullet.projectiles',
 'royalnet.engineer.pda',
 'royalnet.engineer.pda.extensions',
 'royalnet.engineer.pda.implementations',
 'royalnet.lazy',
 'royalnet.lazy.tests',
 'royalnet.scrolls',
 'royalnet.sculptor']

package_data = \
{'': ['*']}

install_requires = \
['async-property>=0.2.1,<0.3.0',
 'pydantic>=1.8.1,<2.0.0',
 'sqlalchemy>=1.4.5,<2.0.0',
 'toml>=0.10.1,<0.11.0']

setup_kwargs = {
    'name': 'royalnet',
    'version': '6.4.0',
    'description': 'A multipurpose bot framework',
    'long_description': '# `royalnet` 6\n\nThe repository for the development of `royalnet` version `6.0.0` and later.\n\nThe documentation is [hosted on Read The Docs](https://royalnet-6.readthedocs.io/en/latest/).\n\n## See also\n\n### PDA implementations\n\n- [royalnet-discordpy](https://github.com/Steffo99/royalnet-discordpy) (based on a Discord Bot)\n- [royalnet-console](https://github.com/Steffo99/royalnet-console) (based on a terminal session)\n\n### Old Royalnet versions\n\n- [Royalnet 5](https://github.com/Steffo99/royalnet-5)\n- [Royalnet 4](https://github.com/Steffo99/royalnet-5/tree/four)\n- [Royalbot 3](https://github.com/Steffo99/royalbot-3)\n- [Royalbot 2](https://github.com/Steffo99/royalbot-2)\n- [Royalbot 1](https://github.com/Steffo99/royalbot-1)\n',
    'author': 'Stefano Pigozzi',
    'author_email': 'me@steffo.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Steffo99/royalnet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import abc

from .. import casing


class BulletContents(casing.Casing, metaclass=abc.ABCMeta):
    """
    Abstract base class for bullet contents.
    """


__all__ = (
    "BulletContents",
)



from __future__ import annotations
from ._imports import *

import datetime

if t.TYPE_CHECKING:
    from .channel import Channel
    from .user import User
    from .button_reaction import ButtonReaction


class Message(BulletContents, metaclass=abc.ABCMeta):
    """
    An abstract class representing a chat message.
    """

    @ap.async_property
    async def text(self) -> t.Optional[str]:
        """
        :return: The raw text contents of the message.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def timestamp(self) -> t.Optional[datetime.datetime]:
        """
        :return: The :class:`datetime.datetime` at which the message was sent.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def reply_to(self) -> t.Optional[Message]:
        """
        :return: The :class:`.Message` this message is a reply to.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def channel(self) -> t.Optional["Channel"]:
        """
        :return: The :class:`.Channel` this message was sent in.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def sender(self) -> t.Optional["User"]:
        """
        :return: The :class:`.User` who sent this message.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def files(self) -> t.Optional[t.List[t.BinaryIO]]:
        """
        :return: A :class:`list` of files attached to the message.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def reactions(self) -> t.List["ButtonReaction"]:
        """
        :return: A :class:`list` of reaction buttons attached to the message.
        """

    async def reply(self, *,
                    text: str = None,
                    files: t.List[t.BinaryIO] = None) -> t.Optional[Message]:
        """
        Reply to this message in the same channel it was sent in.

        :param text: The text to reply with.
        :param files: A :class:`list` of files to attach to the message. The file type should be detected automatically
                      by the frontend, and sent in the best format possible (if all files are photos, they should be
                      sent as a photo album, etc.).
        :return: The sent reply message.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Message",
)



from __future__ import annotations
from ._imports import *

if t.TYPE_CHECKING:
    from .message import Message
    from .user import User


class Channel(BulletContents, metaclass=abc.ABCMeta):
    """
    An abstract class representing a channel where messages can be sent.
    """

    @ap.async_property
    async def name(self) -> t.Optional[str]:
        """
        :return: The name of the message channel, such as the chat title.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def topic(self) -> t.Optional[str]:
        """
        :return: The topic (description) of the message channel.
        """
        raise exc.NotSupportedError()

    @ap.async_property
    async def users(self) -> t.List["User"]:
        """
        :return: A :class:`list` of :class:`.User` who can read messages sent in the channel.
        """
        raise exc.NotSupportedError()

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> t.Optional["Message"]:
        """
        Send a message in the channel.

        :param text: The text to send in the message.
        :param files: A :class:`list` of files to attach to the message. The file type should be detected automatically
                      by the frontend, and sent in the best format possible (if all files are photos, they should be
                      sent as a photo album, etc.).
        :return: The sent message.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Channel",
)



"""
This module contains the base :class:`.Casing` class.
"""

from __future__ import annotations

import abc


class Casing(metaclass=abc.ABCMeta):
    """
    :class:`.Casing`\\ s are parts of the data model that :mod:`royalnet.engineer` uses to build a common interface
    between different applications (*PDA implementations*).

    They use :func:`~async_property.async_property` to represent data, as it may be required to fetch it from a remote
    location before it is available.

    **All** their methods can have three different results:

    - :exc:`.exc.CasingException` is raised, meaning that something went wrong during the data retrieval.
      - :exc:`.exc.NotSupportedError` is raised, meaning that the frontend does not support the feature the requested
        data is about (asking for :meth:`~royalnet.engineer.bullet.contents.message.Message.reply_to` in an SMS
        implementation, for example).

    - :data:`None` is returned, meaning that there is no data in that field (if a message is not a reply to anything,
      :meth:`Message.reply_to` will be :data:`None`.

    - The data is returned.
    """

    def __init__(self):
        """
        Instantiate a new instance of this class.
        """

    @abc.abstractmethod
    def __hash__(self) -> int:
        """
        :return: A :class:`int` value that uniquely identifies the object in this Python interpreter process.
        """
        raise NotImplementedError()
